// alloc 1024 memory blocks and test for errors
//   free good memory blocks (for other applications)
//   block bad memory blocks by sleeping forever after freeing good ones
// if some bits are bad, 0.1% of memory will be occupied and rest can be used
// alternativ: linux badram/badmem kernel-option
//
// gcc -Wall -O2 -fopenmp -o mem_badram mem_badram.c
//     # ToDo:  -funroll-loops -ftree-vectorize
//     # intel-cc: -openmp
// GOMP_CPU_AFFINITY="0-15"  # works as expected (gomp-1,comp2)
// OMP_PROC_BIND=TRUE # thread bounding, thread affinity (dont work!?)
// OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./mem_badram 7168 # 7GB test
// ./mem_badram 31000  # 31GB test, argument1 is physical memory in MB
//
// 2015-10-28 change OMP-loop-vars to signed (avoid pathcc.omp warn+crash)
// 2015-09-04 better avoid in cache-loops, correct speed if malloc-fail
// 2015-08-31 +openmp, fix use of lrnd, +output speed
// 2013-03-22 improved version
// 2011-07-06 cosmetic changes for output (more info)
// 2013 better use memspeed --test=8 or 16  ???
//
// ToDo: output option, -q quiet for multi-node-mode (getenv SLURM),
//       repeat for nh houres, memcpy or vector statt single copy?
//  OMP_SCHEDULE=dynamic,1 # default
//  OMP_SCHEDULE=static,1  # test
// ToDo: slurm shared? sample job-script
// ToDo: disable ECC in BIOS for tests? Asus=yes(ibio)
//       lsmod | grep edac # sb_edac edac_core
//       cat /sys/devices/system/edac/mc/mc[01]/[uc]e_*count|xargs
//       https://www.kernel.org/doc/Documentation/edac.txt
//   dmidecode 64bit = internal ECC only (if so)
//             72bit = external ECC on Board or CPU(+memctl) or FakeECC 
//    k2.6.16 /sys/devices/system/machinecheck/machinecheck0/bank* # mcelog?
//      
// ToDo: drop cashes before start, lock mempage (to shm-file?)
// ToDo: spinpack: Immunity-aware programming, p.e. check sum matrix_elements
//     and compare every iteration (stored=memerror, recompute=compute_error?)
//
// before that do: echo 3 >/proc/sys/vm/drop_caches
// use: dmidecode # to detect ECC-functionality
// GOMP_CPU_AFFINITY="0-15:1" 
// 2015-09x31_______2GB_rw1.5GB                 348MB/s 1t 2004 noECC PentiumM-1400MHz@1200MHz gcc451
// 2015-09ibioN_2x2x1GB_rw900MB/c  2.8GB/s 4t  1030MB/s 1t 2006 ECC Opt2214-2.2GHz gcc412-O2+unroll+vect dmidecode=64bit-DIMMs board=SbECC (aber 9chips/DIMM) 2*2of4DIMMs gcc492+k8=2.5GB/s/4c bios.ecc=on memtest86+v420=2GB/s 4*singletask=4*720MB/s
// 2015-09ibioM_1x4x1GB_rw1.6GB/c  1.4GB/s 2t  1168MB/s 1t 2006 ECC Opt275-2.2GHz  gcc412-O2+unroll+vect dmidecode=72bit-DIMMs
// 2015-09mkautz12x2x2GB_rw500MB/c 876MB/s 6t   162MB/s 1t 2008 ECC m5k.gcc412 -O2 unroll no_openmp! 6t=6*single
// 2015-10mkautz12x2x2GB_rw500MB/c 6xxMB/s 6t   162MB/s 1t 2008 ECC m5k.gcc424 -O2 unroll +openmp t6schwankt=413-690MB/s! 2t=303MB/s 3t=420MB/s 4t=bad=290-305
// 2015-10mkautz12x2x2GB_rw500MB/c 780MB/s 6t   162MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp loop-var-ijk-must-be-signed!       3t=409MB/s 500MHz
// 2015-10asgard12x2x2GB_rw500MB/c 929MB/s 6t   203MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp                                    3t=505MB/s 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c      -       202MB/s 1t 2008 ECC m5k.gcc412=gcc464 -O2 unroll 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c 1.0GB/s 6t   198MB/s 1t 2008 ECC m5k.gcc464 -O2 unroll +openmp 680MHz mit/ohne bootstrap          3t=569MB/s 6t=6*1t-20%
// 2015-10asgard12x2x2GB_rw500MB/c 899GB/s 6t   170MB/s 1t 2008 ECC m5k.gcc492 -O2 unroll +openmp 680MHz --disable-bootstrap
// 2015-10mkautzsp_1x1GB_rw500MB/c      -       902MB/s 1t 2008 noECC?mc! Sempron3500+800MHz k8=sse2+3dnow
// 2015-09quantum1x4x2GB_rw1GB/c:  1.7GB/s 4t   1.2GB/s 1t 2008 ECC E5405-2GHz 228W-200W dmidecode=72bit-DIMMs DDR2 Multi-bit-ECC
// 2015-09quantum2x4x4GB_rw2GB/c:  2.3GB/s 8t   643MB/s 1t 2008 ECC E5450-3GHz
// 2015-09jsPC_____1x2GB_rw1GB     1.2GB/s 2t   1.2GB/s 1t 2010 noECC E5400-2.7GHz@1.6GHz
// 2015-09einstein2x2GB_rw1.5GB    2.7GB/s 4t   2.4GB/s 1t 2010 ECC? 4c8t-i7-860-2.8GHz gcc44-32bit! no-sse4-used! dmidecode=72bit-DIMMs BoardECC=none
// 2015-09landau__6x4GB_rw12GB     5.3GB/s 4t   4.1GB/s 1t 2010 ECC? 4c*2HT-iCore_i7-950-3.06GHz-8MB SSE4.2 24GB dmidecode=2304/2244bits? board=noECC
// 2015-08comp2__4x4x4GB_rw3GB/c: 15.8GB/s 16t  2.0GB/s 1t 2011 ECC Opt6134-2.3GHz
// 2015-09comp2n_4x4x8GB_rw7GB/c: 16.5GB/s 16t  2.1GB/s 1t 2011 ECC Opt6134-2.3GHz.gcc492-O2+unroll+vect
// 2015-09gibbs____4x4GB_rw14GB    5.9GB/s 4t   5.1GB/s 1t 2011 noECC? i5-2400-3.1GHz gcc412-O2+unroll+vect+omp hdc=30130h=3.4y dmidecode=64bit-DIMMs board=noECC
// 2015-09hpmini_2GB_rw1500MB:    1020MB/s 2t   990MB/s 1t 2012 noECC N455-1660MHz+gcc472-O2+unroll+vec BitError!
// 2015-09meitner__4x8GB_rw16GB    5.6GB/s 4t   5.3GB/s 1t 2013 noECC? 4c8ht-i7-3770-Gen2-3410MHz-avx DDR3 dmidecode=64bit-DIMMs
// 2015-09dyson__1x4x8GB_rw7GB/c:  6.8GB/s 4t   6.0GB/s 1t 2014 noECC i5-4460-3.2GHz+gcc482-25.6GB/s 80W dmidecode=64bit-DIMMs
// 2015-09dyson__1x4x8GB_rw7GB/c:  5.9GB/s 4t   2.1GB/s 1t 2014 noECC i5-4460-800MHz+gcc482 49W
// 2015-09msiKS____2x4GB_rw4GB     5.6GB/s 4t   5.1GB/s 1t 2015 noECC i5-4210H-2.9GHz+gcc492-O3+unroll+vec
/* --- 2015-09hpmini_2GB_2012 bit-flip-error ---
 malloc 1024 MiB (Mi=2^20) long= 8 B = 64 bit (arg1)
 malloc 1024 blocks to 1024 kiB = 131072 * long (ki=2^10)
 starting, only every 128th block shown ...
 a=00007fb978db0010 = 130789 GB blk    0 (alloc+ini)
 a=00007fb9708a0010 = 130789 GB blk  128 (alloc+ini)
  ...
 a=00007fb948620010 = 130789 GB blk  768 (alloc+ini)
 a=00007fb9405a0010 = 130789 GB blk  896 (alloc+ini)
 a=00007fb978db0010 iLoop=00/16 ffffffffffffffff
 a=00007fb978db0010 iLoop=02/16 ffffffffffffffff
 loop=  0 speed(w+r):        887 MB/s (1500.000 MB)
 a=00007fb957361010 iLoop=00/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=02/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=04/16 9bc3679699ccb78c  # 1-to-0 bit 15
ERR bad data blk=734 lrnd= 9bc3679699ccb78c  sum=00008000 (j=4) on ?
ERR bad data   a[0001c603]=9bc3679699cc378c diff=00008000 (j=4)
ERR bad data blk=734 lrnd= d5176bb59daaa713  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=d5176bb59daa2713 diff=00008000 (j=2)
ERR bad data blk=734 lrnd= ffffffffffffffff  sum=00008000 (j=8) on ?
ERR bad data   a[0001c603]=ffffffffffff7fff diff=00008000 (j=8)
ERR bad data blk=734 lrnd= ee6f32189910bc72  sum=00008000 (j=10) on ?
ERR bad data   a[0001c603]=ee6f321899103c72 diff=00008000 (j=10)
ERR bad data blk=734 lrnd= ef6306d1f2128d9c  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=ef6306d1f2120d9c diff=00008000 (j=2)
ERR bad data blk=734 lrnd= 7fdcc2331befd79f  sum=00008000 (j=1) on ?
ERR bad data   a[0001c603]=7fdcc2331bef579f diff=00008000 (j=1)
ERR bad data blk=734 lrnd= be583b369497f865  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=be583b3694977865 diff=00008000 (j=2)
 haeufig, dann weniger, dann haeufiger bei parallelem 20MB-scan = 20kB blks
 heating-related? ToDo: how to lock? as /dev/shm-file?
 ext2 did crash spontaniosly, so I did the memory test
 addr changing from time to time
 memtest: DDR2-333 2GB 2GB/s
   error: 00023e8d023 - 574.5 MB  2015-09-09 kernel badram/badmem-fix?
  locking survives s2ram + s2dsk = suspend-to-disk = hibernate (2 tests)
*/      
//  https://en.wikipedia.org/wiki/ECC_memory
//    spacecraft Cassini–Huygens 1997 2.5Gbit-ECC 280-single-bit-errors/day
//               on day 1997-11-06 4xErrorRate (solar-event)
//    pubs 2007-2009 1e-10..1e-17 errors/bit*h = ca. 1err/GB*h or less
//    google: 8% od DIMMs per year, 7e-11err/bit*h or 5err/8GB*h Streuung?
// mem-intense apps may be more efficient on low-clock (or add cpu-intense app, pe. thermo+psu-stress)
// ToDo: add volatile array, correct speed (less cache dependency)
//  gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c
//   ASM? gcc492: unroll8*xmm0(128bit).store + unroll8*rax.cmp
// ToDo: does suspend to disk, reorders phys memory? locking lost?
// ToDo: if found, test time to flip or row-hammer dependence (idle)
//
// bad-mem-successed on:
//  yyyy-mm-dd machine remarks


#include <unistd.h>  /* sleep(600) needs extra h-file */
#include <stdlib.h>  /* system("sleep 600"); */
#include <stdio.h>
#include <sys/time.h>
#ifdef _OPENMP
#include <omp.h>
#endif

typedef unsigned long int ULong;

double fstime(){
 struct timeval tv; if(gettimeofday(&tv, NULL))perror("fstime");
 return tv.tv_sec+1e-6*tv.tv_usec; }
 
#define INVLOOPs 16  /* 2015-08, if 1 gcc492 optimizes loops away ToDo: ??? */

char * dummy_nodename = "?";
char * slurmd_nodename = "??"; /* SLURMD_NODENAME */

// analyse memory given by pointer and length in sizeoflong
//  gcc492 -O2 -funroll-loops ... ToDo: optimiert loops to cache?
//    num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
//    v2015-08: *a=512MB to 512kB-Blocks/core (L2-cache) = Bad (ToDo **a)
//  ToDo: multiple copy mini-array(4kB)
// use volatile to avoid bad optimization, we expect possible memory change
//   problem volatile versus vectorization? check asm-code!
//   gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c | less # ASM-code
// all blocks in INV-loop to avoid in-cache work
int chkmem(ULong volatile ** pa,
    long nb, long num, /* 2D-array params nb=1024(or less on failure) */
    ULong rnd, /* random 64bit-number */
    int procid, /* slurm_proc_id=0..n-1 */
    ULong *bad) { /* marker array [1024] */
  long jj=0 /* return code */, i, k, sum=0; ULong lrnd=rnd; int j, bad_nb=0;
  for (j=0;j<INVLOOPs; j++){ /* fix20150902: why j=1 ... ? bad measure! */
    lrnd=~lrnd;  /* inverse bitpattern between tests */
    if (procid==0 && (j==0 || j==2 || j==4 || j==8)){ /* reduced output */
      printf(" a=%016lx iLoop=%02d/%d %016lx\n", 
        (ULong)pa[0], j, INVLOOPs, lrnd); fflush(stdout);
      /* slurm needs flush to see progress 2015-09-03 */
    }
    for (i=0;i<nb;i++) { /* 1024 blocks */
      ULong volatile * a = pa[i];
      if (pa[i]) {
        // ToDo: vectorize lrnd
#pragma omp parallel for default(shared) private(k) firstprivate(lrnd)
        for (k=0; k<num; k++) a[k]=lrnd; // 2015-08-31 rnd changed to lrnd
        // ToDo: fast scan for error + if error slow scan + show info
        // ToDo: ompi-reduction!
      } // if valid block
    } // block-loop (big data avoids cache intern work) 2015-09-03 2nb-loops
    for (i=0;i<nb;i++) { /* 1024 blocks */
      ULong volatile * a = pa[i];
      if (pa[i]) {
        // ToDo: vectorize lrnd
        sum=0;
#pragma omp parallel for default(shared) private(i)\
  firstprivate(lrnd) reduction(|:sum)
        for (k=0; k<num; k++) sum |= a[k]^lrnd; // 2015-09-03 vectorize +5%
        /* slow on error! */
        if (sum) {
          printf("ERR bad data blk=%4lu lrnd= %016lx  sum=%016lx (j=%u) on %s\n",
            i, lrnd, sum, j /* round */, slurmd_nodename); 
          for (k=0; k<num; k++) if (a[k]!=lrnd) { // better sum|=a[i]^lrnd?
            printf("ERR bad data   a[%08lx]= %016lx diff=%016lx (j=%u)\n",
              k, (ULong)a[k], (a[k]^lrnd), j /* round */); 
            jj=k;
          }
          if (!jj) printf("ERR bad data could not localized\n");
        }
      } // if good block (no errors until yet)
      if (jj/* || sum*/) { bad[i]++; bad_nb++; jj=0; } // count bad blocks + mark as bad
    }  // 1024 mem-blocks i=...
    if (bad_nb) break;
  } // INV-loop
  return bad_nb;
}

int main(int argn, char *argv[]){
  ULong volatile * pa[1024];
  ULong rnd, rnd1, rnd2; long ii, i, j;
  ULong bad[1024];
  long num, nb=0, bad_nb=0, mb=1024; // 2015-10 signed for pathcc openmp
  double t1, t2; char *sprocid=NULL;
  int procid=0; /* slurm_procid */
//  int verbose=1; /* 0=quiet(for multinode) ToDo: check slurm_env! */
#ifdef _OPENMP
  int omp_n= 1;  /* openmp-threads */
#endif
  int sl=sizeof(ULong);            /* sizeLong should be 8 on 64bit or 4 on 32bit */
  if (argn>1) mb=atoi(argv[1]);    /* memory in MB */
  num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
  // SLURM_PROCID  #  0..SLURM_NPROCS
  sprocid=getenv("SLURM_PROCID");  /* string: 0..n or "" */
  slurmd_nodename=getenv("SLURMD_NODENAME");  /* nodename or "" */
  if (!slurmd_nodename)
    slurmd_nodename=getenv("HOSTNAME");  /* nodename or "" CentOS5 */
  if (!slurmd_nodename)
    slurmd_nodename=dummy_nodename; /* we dont want NULL-pointer */
  if (sprocid) procid = atoi(sprocid); /* id0 is talking, id1++ is silent */
  if (procid<3) {
    printf(" SLURM_PROCID=%s = %d (skip3++)\n", sprocid, procid); /* 0..n or "" */
    fflush(stdout);
  }
#ifdef _OPENMP
#pragma omp parallel
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  if(procid==0)printf(" using OpenMP OMP_NUM_THREADS= %d\n", omp_n);
  if(procid==1)printf(" using OpenMP OMP_NUM_THREADS= %d (slurm=1)\n", omp_n);
#endif
  if (!procid) {
    printf(" malloc %lu MiB (Mi=2^20) long= %d B = %d bit (arg1)\n",
           mb /* MB */, sl /* sizeofLong */, sl*8 /* bits */);
    // split into 1024 blocks of size mb*1024
    printf(" malloc 1024 blocks to %lu kiB = %lu * long (ki=2^10)\n", mb, num);
    printf(" starting, only every 128th block shown ...\n"); fflush(stdout);
  }
  for (i=0;i<1024;i++) {
    bad[i] = 0; /* marked as clean */
    pa[i] = (ULong *) malloc(mb*1024l); /* alloc mb*kB (memoryMB/1024) */
    
    if (procid==0 && (i&127)==0)
      printf(" a=%016lx = %3lu GB blk %4lu (alloc+ini)\n", // old: touch
             (ULong)pa[i], ((ULong)pa[i])>>30, i);
    if (!pa[i]) break;
    // touching j+=512 (*8B=4kB) is faster, but gives bad binding on openmp
#pragma omp parallel for default(shared) private(j)
    for (j=0; j<num; j++) pa[i][j]=0; // single task faster if j+=512 (8B)
  } nb=i; // num blocks (serial)
  if (nb!=1024) 
    printf(" WARNING: only %lu blocks = %lu kiB allocated on %s\n",
           nb+1, (nb+1)*mb, slurmd_nodename);
  if (!nb) exit(1);
  for (ii=0;ii<100;ii++) { // 100 rounds
   rnd1=rand();
   rnd2=rand();
   rnd=(rnd1<<32) | rnd2;  // 64bit random
   if ((ii&7)==0) rnd=0;
   // printf(" rnd %016lx %016lx %016lx\n", rnd1, rnd2, rnd);
   t1=fstime();
   bad_nb+=chkmem(pa,nb,num,rnd,procid,bad); // 0=OK 16*invert+write+read+cmp
   t2=fstime();
   if (procid==0 && t1!=t2) /* avoid div0 */
     printf(" loop=%3d speed(w+r): %10.0f MB/s (%.3f MB)\n",
            (int)ii, INVLOOPs*1.*nb/1024.*mb/(t2-t1), mb*1.*nb/1024.);
   if (bad_nb>3) break;
  } // loop100
  for (bad_nb=i=0;i<nb;i++) 
    if (pa[i] && !bad[i]) free((void *)pa[i]); // free good blocks
         else if (bad[i]) bad_nb++; // count bad blocks
  if (bad_nb) {
    printf("%lu bad blocks detected, start sleeping ...\n", bad_nb);
    for (;;) sleep(600); // if(system("sleep 600"))break; // sleep forever
  }
  // ToDo: how to integrate that into the jobsystem? extra-shared-queue?
  return 0;
}
